import torch
from pytorch3d.transforms import random_quaternions

class CameraViewConfig:
    """
    Represents the orientation of a camera in 3D space, defined by a quaternion.
    The class allows for setting a specific quaternion or generating a random one.
    """

    def __init__(self, quaternion: torch.Tensor = None, set_random: bool = False):
        """
        Initializes the CameraViewConfig instance.

        Parameters:
        quaternion (torch.Tensor, optional): The quaternion representing the camera's rotation.
                                             If None, a default or random quaternion is set.
        set_random (bool, optional): If True, sets a random quaternion. Used only if quaternion is None.
                                     Defaults to False.
        """
        if quaternion is not None:
            self.quaternion = quaternion
        elif set_random:
            self.quaternion = random_quaternions(1)
        else:
            # Default quaternion representing no rotation
            self.quaternion = torch.tensor([1, 0, 0, 0])

    def to(self, device) -> 'CameraViewConfig':
        """
        Transfers the quaternion tensor to the specified device (e.g., GPU or CPU).

        Parameters:
        device: The device to transfer the tensor to.

        Returns:
        CameraViewConfig: Returns the instance itself after moving the tensor.
        """
        self.quaternion = self.quaternion.to(device)
        return self