import torch
import mrcfile
import numpy as np

class DensityMapGenerator:
    """
    A class for generating 3D density maps based on Gaussian potentials.
    """


    def __init__(self, grid_span, resolution, device='cuda'):
        """
        Initializes the DensityMapGenerator object.

        Parameters:
        grid_span (float): Defines the half-width of the grid. The grid is cubic and 
                           symmetric around the origin, spanning from -grid_span to +grid_span 
                           in each dimension.
        resolution (int): The number of grid points along each axis. The total number of 
                          points in the grid will be resolution^3.
        device (str, optional): The device where the computations will be performed 
                                ('cuda' for GPU or 'cpu'). Defaults to 'cuda'.

        Attributes:
        self.grid_span (float): Stores the provided grid_span value.
        self.resolution (int): Stores the provided resolution value.
        self.device (torch.device): Represents the device for computation.
        self.grid_points (torch.Tensor): Stores the grid points coordinates in a flattened format.
        """
        self.grid_span = grid_span
        self.resolution = resolution
        self.device = torch.device(device if torch.cuda.is_available() else 'cpu')
        self.grid_points = self._create_grid()

    def _create_grid(self) -> torch.Tensor:
        """
        Creates a 3D grid of points.

        Returns:
        torch.Tensor: A tensor containing the coordinates of the grid points.
        """
        x = torch.linspace(-self.grid_span, self.grid_span, self.resolution, device=self.device)
        y = torch.linspace(-self.grid_span, self.grid_span, self.resolution, device=self.device)
        z = torch.linspace(-self.grid_span, self.grid_span, self.resolution, device=self.device)
        X, Y, Z = torch.meshgrid(x, y, z, indexing='ij')
        return torch.stack([X.flatten(), Y.flatten(), Z.flatten()], dim=-1)

    def create_density_map(self, means: torch.Tensor, sigma: float) -> torch.Tensor:
        """
        Creates a density map based on provided Gaussian parameters.

        Parameters:
        means (torch.Tensor): A tensor of shape (n, 3) representing the means (centers) 
                              of the Gaussian distributions.
        sigma (float): The standard deviation for the Gaussians, assumed isotropic.

        Returns:
        torch.Tensor: The density map as a 3D tensor.
        """
        sigma = torch.eye(3, device=self.device) * sigma
        means = means[:, None, :].to(self.device)
        inv_sigma = torch.linalg.inv(sigma)

        diff = self.grid_points - means
        maha_sq = torch.sum((diff @ inv_sigma) * diff, dim=-1)
        norm_const = torch.sqrt((2 * torch.pi)**3 * torch.linalg.det(sigma))

        gauss_vals = torch.exp(-0.5 * maha_sq) / norm_const
        density_map = torch.sum(gauss_vals, axis=0).view((self.resolution,) * 3)
        return density_map

    def save_density_map(self, density_map: torch.Tensor, filename: str) -> None:
        """
        Saves the density map to a file in MRC format.

        Parameters:
        density_map (torch.Tensor): The density map to be saved.
        filename (str): The name of the file where the density map will be saved.
        """
        density_map_cpu = density_map.cpu().numpy()
        with mrcfile.new(filename, overwrite=True) as mrc:
            mrc.set_data(density_map_cpu.astype(np.float32))
        print(f"Density map saved as {filename}")
