from typing import Optional

import torch
from pytorch3d.transforms import quaternion_apply

class GaussianPotential:
    """
    Represents a Gaussian-based potential field. This class is used to model and manipulate 
    a set of Gaussian distributions, characterized by their means, covariance.
    """

    def __init__(self, means: Optional[torch.Tensor] = torch.zeros(0, 3), sigma: float = 0.1):
        """
        Initializes the GaussianPotential object.

        Parameters:
        means (Optional[torch.Tensor]): A tensor of shape (N, 2 or 3) specifying the mean positions 
                                        of the Gaussians. If not provided, defaults to an empty tensor 
                                        indicating no Gaussians.
        sigma (float): The standard deviation of the Gaussians. Defaults to 0.1.
        """
        self.means = means
        self.sigma = sigma

    def to(self, device):
        """
        Moves the Gaussian means to the specified device.

        Parameters:
        device: The PyTorch device (e.g., 'cpu', 'cuda') to which the Gaussian means are to be moved.

        Returns:
        GaussianPotential: The current instance of GaussianPotential with means moved to the specified device.
        """
        self.means = self.means.to(device)
        return self

    def rotate(self, quaternion) -> 'GaussianPotential':
        """
        Rotates the Gaussians using a quaternion.

        Parameters:
        quaternion: A quaternion representing the rotation to be applied to the Gaussian means.

        Returns:
        GaussianPotential: A new instance of GaussianPotential with the rotated means.
        """
        rotated_means = quaternion_apply(quaternion, self.means)
        return GaussianPotential(means=rotated_means, sigma=self.sigma)

    def project_to_2d(self) -> 'GaussianPotential':
        """
        Projects the 3D Gaussians to 2D.

        This method assumes that the Gaussians are in 3D space and projects them onto a 2D plane
        by ignoring the z-component of the means.

        Returns:
        GaussianPotential: A new instance of GaussianPotential with the means projected to 2D.
        """
        means_2d = self.means[:, :2]
        return GaussianPotential(means=means_2d, sigma=self.sigma)