import torch
import numpy as np
from Bio.PDB import MMCIFParser

class MMCIFParserWrapper:
    """
    A wrapper class for BioPython's MMCIFParser, designed to facilitate extraction of structural
    information from MMCIF (Macromolecular Crystallographic Information File) format files. 
    This class simplifies the process of parsing .cif files and provides convenient methods for
    extracting and manipulating atomic coordinates.

    Attributes:
        parser (MMCIFParser): An instance of MMCIFParser, used for parsing .cif files.
        structure (Structure): A BioPython Structure object, representing the 3D structure 
                               of a molecule as described in the parsed .cif file.
    """

    def __init__(self, datapath: str) -> None:
        """
        Initializes the MMCIFParserWrapper with a specified MMCIF file.

        Parameters:
        datapath (str): The file path to the .cif file that is to be parsed.
        """
        self.parser = MMCIFParser()
        self.structure = self.parser.get_structure('structure_id', datapath)

    def get_atom_coordinates_raw(self) -> torch.Tensor:
        """
        Extracts raw atomic coordinates from the structure.

        Returns:
        torch.Tensor: A tensor containing the raw 3D coordinates of each atom in the structure.
        """
        coords = []
        for model in self.structure:
            for chain in model:
                for residue in chain:
                    for atom in residue:
                        coords.append(atom.get_coord())
        
        coords_array = np.array(coords)
        coords_tensor = torch.tensor(coords_array, dtype=torch.float32)
        return coords_tensor

    def get_atom_coordinates(self) -> torch.Tensor:
        """
        Processes and normalizes the atomic coordinates from the structure.

        This method centralizes the coordinates, fits them within a unit sphere, and returns 
        the normalized atomic positions as a PyTorch tensor. This is useful for standardized 
        processing or visualization.

        Returns:
        torch.Tensor: A tensor of normalized 3D coordinates of each atom in the structure.
        """
        atomic_positions = self.get_atom_coordinates_raw()
        atomic_positions -= torch.mean(atomic_positions, dim=0)
        atomic_positions /= torch.max(torch.norm(atomic_positions, dim=-1))
        return atomic_positions
