import torch

from .gaussian_potential import GaussianPotential
from .camera_view_config import CameraViewConfig

class Renderer:
    """
    Renderer class for orthographic projecting and rendering Gaussian potential in a 2D space.
    """

    def __init__(self, device: torch.device = torch.device("cpu")):
        """
        Initializes the Renderer object.

        Parameters:
        device (torch.device, optional): The computation device to use for rendering. 
                                         Defaults to CPU.
        """
        self.device = device

    def to(self, device: torch.device) -> 'Renderer':
        """
        Sets the device for the renderer.

        Parameters:
        device (torch.device): The computation device to use for rendering.

        Returns:
        Renderer: The Renderer object with updated device.
        """
        self.device = device
        return self

    def render(self, width: int, height: int, potential: GaussianPotential, cam: CameraViewConfig) -> torch.Tensor:
        """
        Renders the Gaussian potential into a 2D image based on the camera view configuration.

        Parameters:
        width (int): The width of the output 2D image.
        height (int): The height of the output 2D image.
        potential (GaussianPotential): The 3D Gaussian potential to be rendered.
        cam (CameraViewConfig): The camera view configuration.

        Returns:
        torch.Tensor: A 2D tensor representing the rendered image.
        """
        # Apply rotation to the Gaussian potential using the camera's quaternion
        rotated_potential = potential.rotate(cam.quaternion)

        # Project the 3D Gaussians to a 2D plane
        projected_potential = rotated_potential.project_to_2d()

        # Transfer means to the current device and retain sigma
        means = projected_potential.means.to(self.device)
        sigma = projected_potential.sigma

        # Generate grid coordinates for the 2D plane
        xx, yy = torch.meshgrid(torch.linspace(-1, 1, height), torch.linspace(-1, 1, width), indexing="ij")
        coords_2d = torch.cat([xx[..., None], yy[..., None]], -1).to(self.device)

        # Calculate distances from each grid point to Gaussian means and render the image
        dists = (coords_2d[..., None, :] - means) ** 2
        rendered_image = torch.sum(torch.exp(-torch.sum(dists, -1) / (2 * sigma**2)), -1)

        return rendered_image